function [G,F] = assemble_2d(p,q,f,w,H,n,int,PAR)
% [G,F] = assemble_2d(p,q,f,w,H,n,int,PAR)
% assembly of the Ritz-Galerkin system
%
% function [P,Px,Py] = p(x,y)
% function Q = q(x,y)
%   functions in the bilinear form of the variational problem
% function fxy = f(x,y) 
%   force 
% function [W,Wx,Wy] = w(x,y)
%   models essential homogeneous boundary conditions
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of the b-splines
% int.c{m1,m2}(:)
%   Gauss coefficients for the subset D_m of grid cell m in D
% int.p{m1,m2}(:,1:2) 
%   Gauss points for the subset D_m of grid cell m in D
% PAR
%   algorithmic parameters
% G(H+n,H+n,2*n+1,2*n+1)
%   Ritz-Galerkin matrix in b-spline format
% F(H+n,H+n)
%   Ritz-Galerkin vector in b-spline format
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nassemble_2d     '); end

% initialization 
G = zeros(H+n,H+n,2*n+1,2*n+1); F = zeros(H+n,H+n);

% loop over grid cells 
work = 0; 
for mx = 1:H, for my = 1:H
  work = work + 100/H^2;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end

  area = sum(int.c{mx,my}); 
  if area > 0
    % values of relevant weighted b-splines for cell m
    px = int.p{mx,my}(:,1); py = int.p{mx,my}(:,2); 
    [B,Bx,By] = bspline_2d(w,n,H,px,py); 
    Pxy = p(px,py); Qxy = q(px,py); Fxy = f(px,py);
 
    % i: indices of b-splines overlapping cell m
    % b-spline i has segment n+1-(i-m) on cell m 
    for ix = mx:mx+n, for iy = my:my+n

      % linear form
      Ix=n+1-ix+mx; Iy=n+1-iy+my;
      F(ix,iy) = F(ix,iy)+int.c{mx,my}'*(Fxy.*B{Ix,Iy});
      for jx=mx:mx+n, for jy=my:my+n

        % bilinear form
        Jx=n+1-jx+mx; Jy=n+1-jy+my;
        G(ix,iy,jx-ix+n+1,jy-iy+n+1) = ... 
          G(ix,iy,jx-ix+n+1,jy-iy+n+1)+int.c{mx,my}'* ... 
          bilinear(Pxy,Qxy,B{Ix,Iy},Bx{Ix,Iy},By{Ix,Iy}, ... 
            B{Jx,Jy},Bx{Jx,Jy},By{Jx,Jy});
      end, end
    end, end
  end
end, end

% diagonal entries for irrelevant b-splines -> 1
% corresponding entries of right side -> 0
maxG = norm(G(:),inf);
diagG = G(:,:,n+1,n+1); 
F(abs(diagG)<=maxG*PAR.tol_zero^2) = 0;
diagG(abs(diagG)<=maxG*PAR.tol_zero^2) = 1; 
G(:,:,n+1,n+1) = diagG;

% sparsity
if PAR.info
  fprintf('\b\b\b\b    \n fraction of nonzero matrix entries: %10.3e\n',sum(G(:)~=0)/(H+n)^4)
end

%%%%% LOCAL FUNCTIONS

function [B,Bx,By] = bspline_2d(w,n,H,px,py)
% values and derivatives of weighted b-splines

% weight function
[W,Wx,Wy] = w(px,py);

% b-splines
px = mod(px*H,1); py = mod(py*H,1); 
[sx,dsx] = b_spline(px,n); [sy,dsy] = b_spline(py,n);
for kx=1:n+1, for ky=1:n+1
  b{kx,ky} = sx(:,kx).*sy(:,ky);
  bx{kx,ky} = dsx(:,kx).*sy(:,ky)*H;
  by{kx,ky} = sx(:,kx).*dsy(:,ky)*H;
end, end

% weighted b-splines for relevant segments
for kx=1:n+1, for ky=1:n+1
  B{kx,ky} = W.*b{kx,ky};
  Bx{kx,ky} = W.*bx{kx,ky}+Wx.*b{kx,ky};
  By{kx,ky} = W.*by{kx,ky}+Wy.*b{kx,ky};
end, end

function GIJ = bilinear(Pxy,Qxy,BI,BIx,BIy,BJ,BJx,BJy)
% values of the bilinear form

GIJ = Pxy.*(BIx.*BJx+BIy.*BJy) + Qxy.*BI.*BJ;

