function [G,F] = assemble_2de(model,E,nu,f,w,H,n,int,PAR)
% [G,F] = assemble_2de(model,E,nu,f,w,H,n,int,PAR)
% assembly of the Ritz-Galerkin system
% 
% model
%   'strain' (plane strain) or 'stress' (plane stress)  
% E 
%   Young modulus
% nu
%   Poisson ratio
% function [F1,F2] = f(x,y)
%   force
% function [W,Wx,Wy] = w(x,y)
%   models essential homogeneous boundary conditions
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of b-splines
% int.c{m1,m2}(:)
%   Gauss coefficients for subset D_m of grid cell m in D
% int.p{m1,m2}(:,1:2) 
%   Gauss points for subset D_m of grid cell m in D
% G(H+n,H+n,2*n+1,2*n+1,2,2)
%   Ritz-Galerkin matrix in b-spline format
% F(H+n,H+n,2)
%   Ritz-Galerkin vector in b-spline format
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nassemble_2de     '); end

% initialization 
switch model
  case 'strain' 
    Q = (E/((1+nu)*(1-2*nu))) * [1-nu; nu; 2-4*nu];
  case 'stress' 
    Q = (E/(1-nu^2)) * [1; nu; 2-2*nu];
end
G = zeros(H+n,H+n,2*n+1,2*n+1,2,2); F = zeros(H+n,H+n,2);

% loop over grid cells 
work = 0;
for mx = 1:H, for my = 1:H
  work = work + 100/H^2;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end
 
  area = sum(int.c{mx,my}); 
  if area > 0
    % values of relevant weighted b-splines for cell m
    px = int.p{mx,my}(:,1); py = int.p{mx,my}(:,2); 
    [B,Bx,By] = bspline_2d(w,n,H,px,py); 
 
    % i: indices of b-splines overlapping cell m
    % b-spline i has segment n+1-(i-m) on cell m 
    for ix = mx:mx+n, for iy = my:my+n

      % linear form
      Ix=n+1-ix+mx; Iy=n+1-iy+my;
      [F1,F2]=f(px,py);      
      FI = int.c{mx,my}' * ([F1,F2].*(B{Ix,Iy}*[1 1]));
      F(ix,iy,1) = F(ix,iy,1) + FI(1);
      F(ix,iy,2) = F(ix,iy,2) + FI(2);
      for jx=mx:mx+n, for jy=my:my+n

        % bilinear form
        Jx=n+1-jx+mx; Jy=n+1-jy+my;
        Sx=jx-ix+n+1; Sy=jy-iy+n+1;
          GIJ = int.c{mx,my}' * bilinear( ... 
            Q,Bx{Ix,Iy},By{Ix,Iy},Bx{Jx,Jy},By{Jx,Jy});
          G(ix,iy,Sx,Sy,1,1) = G(ix,iy,Sx,Sy,1,1)+GIJ(1);
          G(ix,iy,Sx,Sy,2,1) = G(ix,iy,Sx,Sy,2,1)+GIJ(2);
          G(ix,iy,Sx,Sy,1,2) = G(ix,iy,Sx,Sy,1,2)+GIJ(3);
          G(ix,iy,Sx,Sy,2,2) = G(ix,iy,Sx,Sy,2,2)+GIJ(4);
      end, end
    end, end
  end
end, end

% diagonal entries for irrelevant b-splines -> 1 
% corresponding entries of right side -> 0
maxG = norm(G(:),inf);
for m=1:2
  diagG = G(:,:,n+1,n+1,m,m);
  Fm=F(:,:,m);
  Fm(abs(diagG)<=maxG*PAR.tol_zero^2) = 0;
  diagG(abs(diagG)<=maxG*PAR.tol_zero^2) = 1; 
  F(:,:,m)=Fm;
  G(:,:,n+1,n+1,m,m) = diagG;
end

% sparsity
if PAR.info;
  fprintf('\b\b\b\b    \n fraction of nonzero matrix entries: %10.3e\n',sum(G(:)~=0)/(4*(H+n)^4))
end

%%%%% LOCAL FUNCTION

function [B,Bx,By] = bspline_2d(w,n,H,px,py)
% values and derivatives of weighted b-splines

[W,Wx,Wy] = w(px,py);

% update, if Gauss parameters change
px = mod(px*H,1); py = mod(py*H,1); 
[sx,dsx] = b_spline(px,n); [sy,dsy] = b_spline(py,n);
for kx=1:n+1, for ky=1:n+1
  b{kx,ky} = sx(:,kx).*sy(:,ky);
  bx{kx,ky} = dsx(:,kx).*sy(:,ky)*H;
  by{kx,ky} = sx(:,kx).*dsy(:,ky)*H;
end, end

% loop over b-spline segments
for kx=1:n+1, for ky=1:n+1
  B{kx,ky} = W.*b{kx,ky};
  Bx{kx,ky} = W.*bx{kx,ky}+Wx.*b{kx,ky};
  By{kx,ky} = W.*by{kx,ky}+Wy.*b{kx,ky};
end, end

function GIJ = bilinear(Q,BIx,BIy,BJx,BJy)
% entries of I-J block
% 11: (BIx,0,BIy/2) Q (BJx,0,BJy/2)
% 21: (0,BIy,BIx/2) Q (BJx,0,BJy/2)
% 12: (BIx,0,BIy/2) Q (0,BJy,BJx/2)
% 22: (0,BIy,BIx/2) Q (0,BJy,BJx/2)

GIJ = [Q(1)*BIx.*BJx+Q(3)*BIy.*BJy/4, ...
  Q(2)*BIy.*BJx+Q(3)*BIx.*BJy/4, ...
  Q(2)*BIx.*BJy+Q(3)*BIy.*BJx/4, ...
  Q(1)*BIy.*BJy+Q(3)*BIx.*BJx/4]; 

