function [G,F] = assemble_3d(p,q,f,w,H,n,int,PAR)
% [G,F] = assemble_3d(p,q,f,w,H,n,int,PAR)
% assembly of the Ritz-Galerkin system
%
% function [P,Px,Py,Pz] = p(x,y,z)
% function Q = q(x,y,z)
%   functions in the bilinear form of the variational problem
% function fxyz = f(x,y,z)
%   force 
% function [W,Wx,Wy,Wz] = w(x,y)
%   models essential homogeneous boundary conditions
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of b-splines
% int.c{m1,m2,m3}(:)
%   Gauss coefficients for subset D_m of grid cell m in D
% int.p{m1,m2,m3}(:,1:3) 
%   Gauss points for subset D_m of grid cell m in D
% PAR
%   algorithmic parameters
% G(H+n,H+n,H+n,2*n+1,2*n+1,2*n+1)
%   Ritz-Galerkin matrix in b-spline format
% F(H+n,H+n,H+n)
%   Ritz-Galerkin vector in b-spline format
% -> see documentation for details 


% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info; fprintf('\nassemble_3d     '); end

% initialization 
Hn = H+n; nn = 2*n+1;
G = zeros(Hn,Hn,Hn,nn,nn,nn); F = zeros(Hn,Hn,Hn);

% loop over grid cells 
work = 0;
for mx = 1:H, for my = 1:H, for mz=1:H
  work = work + 100/H^3;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end

  area = sum(int.c{mx,my,mz}); 
  if area > 0
    % values of relevant weighted b-splines for cell m
    px = int.p{mx,my,mz}(:,1); 
    py = int.p{mx,my,mz}(:,2); 
    pz = int.p{mx,my,mz}(:,3); 
    [B,Bx,By,Bz] = bspline_3d(w,n,H,px,py,pz); 
    Pxyz = p(px,py,pz); Qxyz = q(px,py,pz); 
    fxyz = f(px,py,pz);
 
    % i: indices of b-splines overlapping cell m
    % b-spline i has segment n+1-(i-m) on cell m 
    for ix=mx:mx+n, for iy=my:my+n, for iz=mz:mz+n

      % linear form
      Ix=n+1-ix+mx; Iy=n+1-iy+my; Iz=n+1-iz+mz;
      F(ix,iy,iz) = F(ix,iy,iz) + int.c{mx,my,mz}' * ... 
        (fxyz.*B{Ix,Iy,Iz}); 
      for jx=mx:mx+n, for jy=my:my+n, for jz=mz:mz+n

        % bilinear form
        Jx=n+1-jx+mx; Jy=n+1-jy+my; Jz=n+1-jz+mz;
        G(ix,iy,iz,jx-ix+n+1,jy-iy+n+1,jz-iz+n+1) = ... 
          G(ix,iy,iz,jx-ix+n+1,jy-iy+n+1,jz-iz+n+1) + ... 
          int.c{mx,my,mz}' * ...
            bilinear(Pxyz,Qxyz,B{Ix,Iy,Iz},Bx{Ix,Iy,Iz}, ... 
            By{Ix,Iy,Iz},Bz{Ix,Iy,Iz},B{Jx,Jy,Jz}, ... 
            Bx{Jx,Jy,Jz},By{Jx,Jy,Jz},Bz{Jx,Jy,Jz});
      end, end, end
    end, end, end
  end
end, end, end

% diagonal entries for irrelevant b-splines -> 1
% corresponding entries of right side -> 0
maxG = norm(G(:),inf);
diagG = G(:,:,:,n+1,n+1,n+1); 
F(abs(diagG)<=maxG*PAR.tol_zero^2) = 0;
diagG(abs(diagG)<=maxG*PAR.tol_zero^2) = 1; 
G(:,:,:,n+1,n+1,n+1) = diagG;

% sparsity
if PAR.info 
  fprintf('\b\b\b\b    \n fraction of nonzero matrix entries: %10.3e\n',sum(G(:)~=0)/(Hn)^6); 
end

%%%%% LOCAL FUNCTION

function [B,Bx,By,Bz] = bspline_3d(w,n,H,px,py,pz)
% values and derivatives of weighted b-splines

% weight function
[W,Wx,Wy,Wz] = w(px,py,pz);

% b-splines
px = mod(px*H,1); py = mod(py*H,1); pz = mod(pz*H,1); 
[sx,dsx] = b_spline(px,n); 
[sy,dsy] = b_spline(py,n);
[sz,dsz] = b_spline(pz,n);
for kx=1:n+1, for ky=1:n+1, for kz=1:n+1
  b{kx,ky,kz} = sx(:,kx).*sy(:,ky).*sz(:,kz);
  bx{kx,ky,kz} = dsx(:,kx).*sy(:,ky).*sz(:,kz)*H;
  by{kx,ky,kz} = sx(:,kx).*dsy(:,ky).*sz(:,kz)*H;
  bz{kx,ky,kz} = sx(:,kx).*sy(:,ky).*dsz(:,kz)*H;
end, end, end

% weighted b-splines for relevant segments
for kx=1:n+1, for ky=1:n+1, for kz=1:n+1
  B{kx,ky,kz} = W.*b{kx,ky,kz};
  Bx{kx,ky,kz} = W.*bx{kx,ky,kz}+Wx.*b{kx,ky,kz};
  By{kx,ky,kz} = W.*by{kx,ky,kz}+Wy.*b{kx,ky,kz};
  Bz{kx,ky,kz} = W.*bz{kx,ky,kz}+Wz.*b{kx,ky,kz};
end, end, end

function GIJ = bilinear(Pxyz,Qxyz,BI,BIx,BIy,BIz,BJ,BJx,BJy,BJz)
% values of the bilinear form

GIJ = Pxyz.*(BIx.*BJx+BIy.*BJy+BIz.*BJz) + Qxyz.*BI.*BJ;

