function [G,F] = assemble_3de(E,nu,f,w,H,n,int,PAR)
% [G,F] = assemble_3de(E,nu,f,w,H,n,int)
% assembly of the Ritz-Galerkin system
%
% E 
%   Young modulus
% nu
%   Poisson ratio
% function fxyz = f(x,y,z)
%   internal force, column of values for each component 
% function [W,Wx,Wy,Wz,Wxx,Wxy,Wxz,Wyy,Wyz,Wzz] = w(x,y,z)
%   models essential homogeneous boundary conditions
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of b-splines
% int.c{m1,m2,m3}(:)
%   Gauss coefficients for subset D_m of grid cell m in D
% int.p{m1,m2,m3}(:,1:3) 
%   Gauss points for subset D_m of grid cell m in D
% G(H+n,H+n,H+n,2*n+1,2*n+1,2*n+1,3,3)
%   Ritz-Galerkin matrix in b-spline format
% F(H+n,H+n,H+n,3)
%   Ritz-Galerkin vector in b-spline format
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nassemble_3de      '); end

% initialization 
Hn = H+n; nn = 2*n+1; 
G = zeros(Hn,Hn,Hn,nn,nn,nn,3,3); F = zeros(Hn,Hn,Hn,3);
lambda = E*nu/((1+nu)*(1-2*nu)); mu = E/(2*(1+nu));

% loop over grid cells 
work = 0;
for mx=1:H, for my=1:H, for mz=1:H
  work = work + 100/H^3;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end

  if sum(int.c{mx,my,mz}) > 0; 
    % values of relevant weighted b-splines for cell m
    px = int.p{mx,my,mz}(:,1); 
    py = int.p{mx,my,mz}(:,2); 
    pz = int.p{mx,my,mz}(:,3); 
    [B,Bx,By,Bz] = bspline_3d(w,n,H,px,py,pz); 
 
    % i: indices of b-splines overlapping cell m
    % b-spline i has segment n+1-(i-m) on cell m 
    for ix=mx:mx+n, for iy=my:my+n, for iz=mz:mz+n
	  
      % linear form
      Ix=n+1-ix+mx; Iy=n+1-iy+my; Iz=n+1-iz+mz;
      [F1,F2,F3]=f(px,py,pz);
      FI = int.c{mx,my,mz}' * ([F1,F2,F3].*(B{Ix,Iy,Iz}*[1 1 1]));
      F(ix,iy,iz,1) = F(ix,iy,iz,1) + FI(1);
      F(ix,iy,iz,2) = F(ix,iy,iz,2) + FI(2);
      F(ix,iy,iz,3) = F(ix,iy,iz,3) + FI(3);

      for jx=mx:mx+n, for jy=my:my+n, for jz=mz:mz+n

        % bilinear form
        Jx=n+1-jx+mx; Jy=n+1-jy+my; Jz=n+1-jz+mz; 
        Sx=jx-ix+n+1; Sy=jy-iy+n+1; Sz=jz-iz+n+1;
        GIJ = int.c{mx,my,mz}' * bilinear(lambda,mu, ... 
            Bx{Ix,Iy,Iz},By{Ix,Iy,Iz},Bz{Ix,Iy,Iz}, ...
            Bx{Jx,Jy,Jz},By{Jx,Jy,Jz},Bz{Jx,Jy,Jz});
        GIJ = reshape(GIJ,[3,3]);
        for k1=1:3, for k2=1:3
            G(ix,iy,iz,Sx,Sy,Sz,k1,k2) = ...
                G(ix,iy,iz,Sx,Sy,Sz,k1,k2)+GIJ(k1,k2);
        end, end
      end, end, end
    end, end, end
  end
end, end, end

% diagonal entries for irrelevant b-splines -> 1
maxG = norm(G(:),inf);
for m=1:3
  diagG = G(:,:,:,n+1,n+1,n+1,m,m);
  Fm=F(:,:,:,m);
  Fm(abs(diagG)<=maxG*PAR.tol_zero^2) = 0;
  diagG(abs(diagG)<=maxG*PAR.tol_zero^2) = 1; 
  F(:,:,:,m)=Fm;
  G(:,:,:,n+1,n+1,n+1,m,m) = diagG;
end

% sparsity
if PAR.info;
  fprintf('\b\b\b\b    \n fraction of nonzero matrix entries: %10.3e\n',sum(G(:)~=0)/(9*(H+n)^6))
end

%%%%% LOCAL FUNCTIONS

function [B,Bx,By,Bz] = bspline_3d(w,n,H,px,py,pz)
% values and derivatives of weighted b-splines

% weight function
[W,Wx,Wy,Wz] = w(px,py,pz);

% univariate b-spline
px = mod(px*H,1); py = mod(py*H,1); pz = mod(pz*H,1);
[sx,dsx] = b_spline(px,n); 
[sy,dsy] = b_spline(py,n);
[sz,dsz] = b_spline(pz,n);

% weighted trivariate b-spline
for kx=1:n+1, for ky=1:n+1, for kz=1:n+1
  b = sx(:,kx).*sy(:,ky).*sz(:,kz) ;
  bx = dsx(:,kx).*sy(:,ky).*sz(:,kz)*H;
  by = sx(:,kx).*dsy(:,ky).*sz(:,kz)*H;
  bz = sx(:,kx).*sy(:,ky).*dsz(:,kz)*H;
  B{kx,ky,kz} = W.*b;
  Bx{kx,ky,kz} = W.*bx+Wx.*b;
  By{kx,ky,kz} = W.*by+Wy.*b;
  Bz{kx,ky,kz} = W.*bz+Wz.*b;
end, end, end

%%%%%

function GIJ = bilinear(lambda,mu,BIx,BIy,BIz,BJx,BJy,BJz)
% entries of I-J block
% (p,q) = 11, 21, 31, 12, 22, 32, 13, 23, 33 
% a( e_p BI , e_q BJ), e unit vectors, B b-splines

% trace(epsilon_I) trace(epsilon_J)
G1 = [BIx.*BJx, BIy.*BJx, BIz.*BJx, ...
      BIx.*BJy, BIy.*BJy, BIz.*BJy, ...
      BIx.*BJz, BIy.*BJz, BIz.*BJz];

% epsilon_I : epsilon_J
S = G1(:,1)+G1(:,5)+G1(:,9);
G2 = G1(:,[1 4 7 2 5 8 3 6 9]);
G2(:,[1 5 9]) = G2(:,[1 5 9])+[S,S,S];

% combine
GIJ = lambda*G1 + mu*G2;

