function [U,wuXY,X,Y,R,e] = elasticity_2d(model,E,nu,f,wD,w,H,n,PAR)
% [U,wuXY,X,Y,R,e] = elasticity_2d(model,E,nu,f,wD,w,H,n,PAR)
%
% Ritz-Galerkin approximation for the Lame-Navier system
%   -div(sigma(u)) = f
% with weighted b-splines
%   u_m ~ sum_k1,k2 U(k1,k2,m) w(x,y) b_k(x,y) (m=1,2)
% on the domain in (0,1)^2 with wD>0
% (plane strain or plane stress model)
%
% model
%   'strain' (plane strain) or 'stress' (plane stress)  
% E 
%   Young modulus
% nu
%   Poisson ratio
% function [F1,F2] = f(x,y)
%   internal force
% function W = wD(x,y)
%   positive on the domain D, subset of (0,1)^2
% function [W,Wx,Wy,Wxx,Wxy,Wzz] = w(x,y)
%   models essential homogeneous boundary conditions
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of the b-splines
% PAR
%   algorithmic parameters
% U(1:H+n,1:H+n,1:2)
%   coefficients of the weighted b-splines
% wuXYZ(1:H*M,1:H*M,1:2)
%  grid values of the weighted spline, M=PAR.values
% X(1:H*M,1:H*M), Y(1:H*M,1:H*M)
%  uniform grid of (H*M)^2 points in (0,1)^2, M=PAR.values
% R(1:M*H,1:M*H,1:2)
%  residuum of the partial differential equation, M=PAR.values 
% e
%   relative error for the differential equation 
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nelasticity_2d\n'); end

% integration, assembly, solution, evaluation, and visualization
int = integrate_2d(wD,H,PAR);
[G,F] = assemble_2de(model,E,nu,f,w,H,n,int,PAR);
U = solve_2de(G,F,PAR);
[wuXY,X,Y,R,e] = evaluate_2de(model,E,nu,f,wD,w,U,H,n,PAR);
if PAR.info > 1, visualize_2de(wD,wuXY,X,Y,H); end

