function [wuXY,X,Y,R,e] = evaluate_2d(p,q,f,wD,w,U,H,n,PAR)
%  [wuXY,X,Y,R,e] = evaluate_2d(p,q,f,wD,w,U,H,n);
%
% evaluation of the solution and the residuum for  
%   -div(p grad u) + q u = f
% for the weighted approximation 
%   u ~ sum_k1,k2 U(k1,k2) w(x,y) b_k(x,y)
% on the domain in (0,1)^2 with wD>0
%
% function [P,Px,Py] = p(x,y)
% function Q = q(x,y)
%   functions in the bilinear form of the variational problem
% function F = f(x,y) 
%   force
% function W = wD(x,y)
%   positive on the domain D, subset of (0,1)^2
% function [W,Wx,Wy] = w(x,y)
%   models essential homogeneous boundary conditions
% U(1:H+n,1:H+n)
%   coefficients of the weighted b-splines
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of the b-splines
% PAR
%   algorithmic parameters
% wUXY(1:H*M,1:H*M)
%    grid values of the weighted spline, M=PAR.values
% X(1:H*M,1:H*M,1:H*M), Y(1:H*M,1:H*M,1:H*M)
%   uniform grid of (H*M)^2 points in (0,1)^2, M=PAR.values
% R(1:H*M,1:H*M,1:H*M)
%   residuum of the differential equation 
% e
%   relative error for the differential equation 
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nevaluate_2d\n'); end 

% weight functions and solution
M=PAR.values;
[X,Y] = ndgrid([1/2:M*H]/(M*H));
WD = wD(X,Y); 
[W,Wx,Wy,Wxx,Wxy,Wyy] = w(X,Y);
VU = spline_2d(U,H,n,M);
wuXY=W.*VU; wuXY(WD<0) = NaN;

if n==1 | sum(isnan(Wxx(:)))>0
  if PAR.info, fprintf(' residual not computable\n'); end
  e = NaN; R = NaN*wuXY; return 
end

% evaluate functions of the pde 
[P,Px,Py] = p(X,Y); Q = q(X,Y); F = f(X,Y);

% values and derivatives of the spline 
VUx = spline_2d(diff(U,1,1)*H,H,[n-1,n],M);
VUy = spline_2d(diff(U,1,2)*H,H,[n,n-1],M); 
VUxx = spline_2d(diff(U,2,1)*H*H,H,[n-2,n],M); 
VUyy = spline_2d(diff(U,2,2)*H*H,H,[n,n-2],M);  

% values and derivatives of the weighted spline 
WUx = Wx.*VU + W.*VUx;
WUy = Wy.*VU + W.*VUy;
WUxx = Wxx.*VU + 2*Wx.*VUx + W.*VUxx;
WUyy = Wyy.*VU + 2*Wy.*VUy + W.*VUyy;

% residuum and error for the pde
R = F + P.*(WUxx+WUyy) + Px.*WUx + Py.*WUy - Q.*wuXY; 
e=R; e(isnan(e)) = 0; F(WD<=0) = 0; 
e = norm(e(:)) / norm(F(:));

if PAR.info, fprintf(' error for the pde: %10.3e\n',e); end

