function [wuXY,X,Y,R,e] = evaluate_2de(model,E,nu,f,wD,w,U,H,n,PAR);
% [wuXY,X,Y,R,e] = error_2de(model,E,nu,f,wD,w,U,H,n,PAR);
%
% evaluation of the solution and the residuum for 
%   -div(sigma(u)) = f
% for the weighted approximation 
%   u_m ~ sum_k1,k2 U(k1,k2,m) w(x,y) b_k(x,y) (m=1,2)
% on the domain in (0,1)^2 with wD>0
% (plane strain or plane stress model)
%
% model
%   'strain' (plane strain) or 'stress' (plane stress)  
% E 
%   Young modulus
% nu
%   Poisson ratio
% function [F1,F2] = f(x,y)
%   internal force
% function W = wD(x,y)
%   positive on the domain D, subset of (0,1)^2
% function [W,Wx,Wy,Wxx,Wxy,Wzz] = w(x,y)
%   models essential homogeneous boundary conditions
% U(1:H+n,1:H+n,1:2)
%   coefficients of the weighted b-splines
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of the b-splines
% PAR
%   algorithmic parameters
% wUXY(1:H*M,1:H*M,1:2)
%    grid values of the weighted spline, M=PAR.values
% X(1:H*M,1:H*M), Y(1:H*M,1:H*M)
%   uniform grid of (H*M)^2 points in (0,1)^2, M=PAR.values
% R(1:H*M,1:H*M,1:2)
%   residuum of the differential equation 
% e
%   relative error for the differential equation 
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nevaluate_2de\n'); end 

% parameters of pde, force and weight functions
switch model
  case 'strain' 
    Q = (E/((1+nu)*(1-2*nu))) * [1-nu; nu; 2-4*nu];
  case 'stress' 
    Q = (E/(1-nu^2)) * [1; nu; 2-2*nu];
end

% weight functions and solution
M=PAR.values;
[X,Y] = ndgrid([1/2:M*H]/(M*H));
[F1,F2] = f(X,Y);
[W,Wx,Wy,Wxx,Wxy,Wyy] = w(X,Y);
VU = spline_2d(U,H,n,M);
WD = wD(X,Y); WD = sign(WD); WD(WD<=0)=NaN;
for m=1:2
  wuXY(:,:,m) = W.*VU(:,:,m).*WD;
end

if n==1 | sum(isnan(Wxx(:)))>0
  if PAR.info, fprintf(' residual not computable\n'); end
  e = NaN; R = NaN*wuXY; return 
end

% values and derivatives of the spline 
VUx = spline_2d(diff(U,1,1)*H,H,[n-1,n],M); 
VUy = spline_2d(diff(U,1,2)*H,H,[n,n-1],M); 
VUxx = spline_2d(diff(U,2,1)*H*H,H,[n-2,n],M); 
VUxy = spline_2d(diff(diff(U,1,1),1,2)*H*H,H,[n-1,n-1],M); 
VUyy = spline_2d(diff(U,2,2)*H*H,H,[n,n-2],M);  

% values and derivatives of the weighted spline 
for m=1:2
  WUx(:,:,m)  = Wx.*VU(:,:,m) + W.*VUx(:,:,m);
  WUy(:,:,m)  = Wy.*VU(:,:,m) + W.*VUy(:,:,m);
  WUxx(:,:,m) = Wxx.*VU(:,:,m) + 2*Wx.*VUx(:,:,m) + W.*VUxx(:,:,m);
  WUxy(:,:,m) = Wxy.*VU(:,:,m) + Wx.*VUy(:,:,m) + Wy.*VUx(:,:,m) + W.*VUxy(:,:,m);
  WUyy(:,:,m) = Wyy.*VU(:,:,m) + 2*Wy.*VUy(:,:,m) + W.*VUyy(:,:,m);
end

% residuum and error for the pde
R(:,:,1) = F1 + Q(1)*WUxx(:,:,1) + ...
  Q(2)*WUxy(:,:,2) + (Q(3)/4)*(WUyy(:,:,1)+WUxy(:,:,2));
R(:,:,2) = F2 + Q(1)*WUyy(:,:,2) + ...
  Q(2)*WUxy(:,:,1) + (Q(3)/4)*(WUxx(:,:,2)+WUxy(:,:,1));

WD(isnan(WD))=0;
e = R(:,:,1).*WD;e(:,:,2) = R(:,:,2).*WD;
F1=F1.*WD;F2=F2.*WD;
e = norm(e(:)) / norm([F1(:);F2(:)]);

if PAR.info, fprintf(' error of pde: %10.3e\n',e); end

