function [wuXYZ,X,Y,Z,R,e] = evaluate_3d(p,q,f,wD,w,U,H,n,PAR);
% [wuXYZ,X,Y,Z,R,e] = error_3d(p,q,f,wD,w,U,H,n,PAR);
%
% evaluation of the solution and the residuum for
%   -div(p grad u) + q u = f
% for the weighted approximation 
%   u ~ sum_k1,k2,k3 U(k1,k2,k3) w(x,y,z) b_k(x,y,z)
% on the domain in (0,1)^3 with wD>0
%
% function [P,Px,Py,Pz] = p(x,y,z)
% function Q = q(x,y,z)
%   functions in the bilinear form of the variational problem
% function F = f(x,y,z) 
%   force
% function W = wD(x,y,z)
%   positive on domain D, subset of (0,1)^3
% function [W,Wx,Wy,Wz,Wxx,Wxy,Wxz,Wyy,Wyz,Wzz] = w(x,y,z)
%   models essential homogeneous boundary conditions
% U(1:H+n,1:H+n,1:H+n)
%   coefficients of the weighted b-splines
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of the b-splines
% PAR
%   algorithmic parameters
% wUXYZ(1:H*M,1:H*M,1:H*M)
%    grid values of the weighted spline, M=PAR.values
% X(1:H*M,1:H*M,1:H*M), Y(1:H*M,1:H*M,1:H*M), Z(1:H*M,1:H*M,1:H*M)
%   uniform grid of (H*M)^3 points in (0,1)^3, M=PAR.values
% R(1:H*M,1:H*M,1:H*M)
%   residuum of the differential equation 
% e
%   relative error for the differential equation 
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nevaluate_3d\n'); end 

% weight functions and solution
M = PAR.values;
[X,Y,Z] = ndgrid([1/2:H*M]/(H*M));
WD = wD(X,Y,Z); 
[W,Wx,Wy,Wz,Wxx,Wxy,Wxz,Wyy,Wyz,Wzz] = w(X,Y,Z);
VU=spline_3d(U,H,n,M);
wuXYZ = W.*VU; wuXYZ(WD<0) = NaN;

if n==1 | sum(isnan(Wxx(:)))>0
  if PAR.info, fprintf(' residual not computable\n'); end
  e = NaN; R = NaN*ones(size(X)); return 
end

% evaluate functions of the pde 
[P,Px,Py,Pz] = p(X,Y,Z); Q = q(X,Y,Z); F = f(X,Y,Z);

% values and derivatives of the spline 
VUx = spline_3d(diff(U,1,1)*H,H,[n-1,n,n],M);
VUy = spline_3d(diff(U,1,2)*H,H,[n,n-1,n],M); 
VUz = spline_3d(diff(U,1,3)*H,H,[n,n,n-1],M); 
VUxx = spline_3d(diff(U,2,1)*H*H,H,[n-2,n,n],M);
VUyy = spline_3d(diff(U,2,2)*H*H,H,[n,n-2,n],M);  
VUzz = spline_3d(diff(U,2,3)*H*H,H,[n,n,n-2],M);  

% values and derivatives of the weighted spline 
WUx = Wx.*VU + W.*VUx;
WUy = Wy.*VU + W.*VUy;
WUz = Wz.*VU + W.*VUz;
WUxx = Wxx.*VU + 2*Wx.*VUx + W.*VUxx;
WUyy = Wyy.*VU + 2*Wy.*VUy + W.*VUyy;
WUzz = Wzz.*VU + 2*Wz.*VUz + W.*VUzz;

% residuum and error for the pde
R = F + P.*(WUxx+WUyy+WUzz) + Px.*WUx+Py.*WUy+Pz.*WUz - Q.*wuXYZ; 
e=R; e(isnan(e)) = 0; F(WD<=0) = 0; 
e = norm(e(:)) / norm(F(:));

if PAR.info, fprintf(' error of pde: %10.3e\n',e); end


