function [wuXYZ,X,Y,Z,R,e] = evaluate_3de(E,nu,f,wD,w,U,H,n,PAR);
% [wuXYZ,X,Y,Z,R,e] = evaluate_3de(E,nu,f,wD,w,U,H,n,PAR);
%
% Ritz-Galerkin approximation of the Lame-Navier system
%   -div(sigma(u)) = f
% with weighted b-splines
%   u_m ~ sum_k U(k1,k2,k3,m) w(x,y,z) b_k(x,y,z) (m=1,2,3)
% on the domain D in (0,1)^3 with wD > 0
%
% E 
%   Young modulus
% nu
%   Poisson ratio
% function F = f(x,y,z)
%   internal force, column of values for each component
% function W = wD(x,y,z)
%   positive on domain D, subset of (0,1)^3
% function [W,Wx,Wy,Wz,Wxx,Wxy,Wxz,Wyy,Wyz,Wzz] = w(x,y,z)
%   models essential homogeneous boundary conditions
% U(1:H+n,1:H+n,1:H+n,1:3)
%   coefficients of weighted b-splines
% H
%   number of cells per coordinate direction, width 1/H
% n 
%   degree of b-splines
% PAR
%   algorithmic parameters
% wUXY(1:H*M,1:H*M,1:H*M,1:3)
%    grid values of the weighted spline, M=PAR.values
% X(1:H*M,1:H*M,1:H*M), Y(1:H*M,1:H*M,1:H*M), Z(1:H*M,1:H*M,1:H*M)
%   uniform grid of (H*M)^3 points in (0,1)^3, M=PAR.values
% R(1:H*M,1:H*M,1:H*M,1:3)
%   residuum of the differential equation 
% e
%   relative error for the differential equation 
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nevaluate_3de\n'); end 

% parameters of pde, force and weight functions
lambda = E*nu/((1+nu)*(1-2*nu));
mu = E/(2*(1+nu));
M=PAR.values;
[X,Y,Z] = ndgrid([1/2:M*H]/(M*H));
[F1,F2,F3] = f(X,Y,Z);
[W,Wx,Wy,Wz,Wxx,Wxy,Wxz,Wyy,Wyz,Wzz] = w(X,Y,Z);
VU = spline_3d(U,H,n,M);
WD = wD(X,Y,Z); WD = sign(WD); WD(WD<=0)=NaN;
for m=1:3
  wuXYZ(:,:,:,m) = W.*VU(:,:,:,m).*WD;
end

if n==1 | sum(isnan(Wxx(:)))>0
  if PAR.info, fprintf(' residual not computable\n'); end
  e = NaN; R = NaN*wuXYZ; return 
end

% values and derivatives of spline 
VUx = spline_3d(diff(U,1,1)*H,H,[n-1,n,n],M); 
VUy = spline_3d(diff(U,1,2)*H,H,[n,n-1,n],M); 
VUz = spline_3d(diff(U,1,3)*H,H,[n,n,n-1],M); 
VUxx = spline_3d(diff(U,2,1)*H*H,H,[n-2,n,n],M); 
VUxy = spline_3d(diff(diff(U,1,1),1,2)*H*H,H,[n-1,n-1,n],M); 
VUxz = spline_3d(diff(diff(U,1,1),1,3)*H*H,H,[n-1,n,n-1],M); 
VUyy = spline_3d(diff(U,2,2)*H*H,H,[n,n-2,n],M);  
VUyz = spline_3d(diff(diff(U,1,2),1,3)*H*H,H,[n,n-1,n-1],M); 
VUzz = spline_3d(diff(U,2,3)*H*H,H,[n,n,n-2],M);  


% values and derivatives of weighted spline 
for m=1:3
  WU(:,:,:,m) = W.*VU(:,:,:,m);
  WUx(:,:,:,m) = Wx.*VU(:,:,:,m) + W.*VUx(:,:,:,m);
  WUy(:,:,:,m) = Wy.*VU(:,:,:,m) + W.*VUy(:,:,:,m);
  WUz(:,:,:,m) = Wz.*VU(:,:,:,m) + W.*VUz(:,:,:,m);
  WUxx(:,:,:,m) = Wxx.*VU(:,:,:,m) + 2*Wx.*VUx(:,:,:,m) + W.*VUxx(:,:,:,m);
  WUxy(:,:,:,m) = Wxy.*VU(:,:,:,m) + Wx.*VUy(:,:,:,m) + Wy.*VUx(:,:,:,m) + W.*VUxy(:,:,:,m);
  WUxz(:,:,:,m) = Wxz.*VU(:,:,:,m) + Wx.*VUz(:,:,:,m) + Wz.*VUx(:,:,:,m) + W.*VUxz(:,:,:,m);
  WUyy(:,:,:,m) = Wyy.*VU(:,:,:,m) + 2*Wy.*VUy(:,:,:,m) + W.*VUyy(:,:,:,m);
  WUyz(:,:,:,m) = Wyz.*VU(:,:,:,m) + Wy.*VUz(:,:,:,m) + Wz.*VUy(:,:,:,m) + W.*VUyz(:,:,:,m);
  WUzz(:,:,:,m) = Wzz.*VU(:,:,:,m) + 2*Wz.*VUz(:,:,:,m) + W.*VUzz(:,:,:,m);
end

% error of pde
% -mu(uk_11+uk_22+uk_33)-(lambda+mu)(u1_1k+u2_2k+u3_3k)=fk
% (k=1,2,3)
R(:,:,:,1) = F1 + ...
  mu*(WUxx(:,:,:,1)+WUyy(:,:,:,1)+WUzz(:,:,:,1)) + ...
  (lambda+mu)*(WUxx(:,:,:,1)+WUxy(:,:,:,2)+WUxz(:,:,:,3)); 
R(:,:,:,2) = F2 + ...
  mu*(WUxx(:,:,:,2)+WUyy(:,:,:,2)+WUzz(:,:,:,2)) + ...
  (lambda+mu)*(WUxy(:,:,:,1)+WUyy(:,:,:,2)+WUyz(:,:,:,3)); 
R(:,:,:,3) = F3 + ...
  mu*(WUxx(:,:,:,3)+WUyy(:,:,:,3)+WUzz(:,:,:,3)) + ...
  (lambda+mu)*(WUxz(:,:,:,1)+WUyz(:,:,:,2)+WUzz(:,:,:,3)); 

WD(isnan(WD))=0;
e = R(:,:,:,1).*WD;e(:,:,:,2) = R(:,:,:,2).*WD;e(:,:,:,3) = R(:,:,:,3).*WD;
F1=F1.*WD;F2=F2.*WD;F3=F3.*WD;
e = norm(e(:)) / norm([F1(:);F2(:);F3(:)]);

if PAR.info, fprintf(' error of pde: %10.3e\n',e); end


