function [int,cells] = integrate_2d(wD,H,PAR)
% [int,cells] = integrate_2d(wD,H,PAR)
% parameters for numerical integration over grid cells 
%
% function [W,Wx,Wy] = wD(x,y)
%   positive on the domain D, subset of (0,1)^2
% H
%   number of cells per coordinate direction, width 1/H
% PAR
%   algorithmic parameters
% int.c{mx,my}(:)
%   Gauss coefficients for the subset D_m of grid cell m in D
% int.p{mx,my}(:,1:2) 
%   Gauss points for the subset D_m of grid cell m in D
% cells(1:H,1:H)
%   type of cells; inner=1, boundary=0, outer=-1 
% -> see documenttion for details 

% authors/copyright 2011: Klaus Hoellig and Joerg Hoerner
% FEMB-Programs (Version: 1.0.1, October 2014)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nintegrate_2d     '); end

% auxiliary parameters and initialization
load parameters_2d
GAUSS2 = GAUSS2{PAR.gauss}; PTS = PTS{PAR.fit}; POL = POL{PAR.fit};
int.c = cell(H,H); int.p = cell(H,H); cells = -ones(H,H);

% loop over grid cells
work = 0;
for mx=1:H, for my=1:H
  work = work+100/H^2;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end
 
  % interpolate values of wD in Q_m on regular grid X,Y -> 
  % Taylor coefficients A with scaled variables in [-1/2,1/2]^2
  A = wD((mx-1+PTS.X)/H,(my-1+PTS.Y)/H);
  A = reshape(POL.V2*A,PAR.fit,PAR.fit);

  % dominance of center value:
  % 2 |A(1,1)| > Aabs = sum_i,j |A(i,j)| 2^(-i-j+2)
  % -> sufficient criterium for excluding boundary cell
  Aabs = abs(A(:))'*POL.E; 
  if 2*abs(A(1,1)) > Aabs
    if A(1,1) > 0 % inner cell 
      int.c{mx,my} = GAUSS2.C/H^2; 
      int.p{mx,my} = [mx-1+GAUSS2.P(:,1),my-1+GAUSS2.P(:,2)]/H;
    else % outer cell
      int.c{mx,my} = 0; 
      int.p{mx,my} = [mx-1/2, my-1/2]/H;
    end
  else
 
    % order of integration, determined by gradient
    % smallest component last -> outer integration (y-index) 
    [g,ind] = sort(-abs([A(2,1), A(1,2)]));
    A = permute(A,ind); 

    % Gauss parameters in [-1/2,1/2]^2
    [cxy,pxy] = int_2d(A,H,PAR,GAUSS,POL);

    % reverse permutation due to order of integration
    pxy(:,ind) = pxy;
    
    % delete entries with weight 0
    nonzero_ind=find(cxy>0);
    cxy=cxy(nonzero_ind);
    pxy=pxy(nonzero_ind,:);

    if isempty(cxy)
      cxy=0;pxy=[0,0];
    end                
    
    % assign parameters to cell 
    int.c{mx,my} = cxy/H^2;
    int.p{mx,my} = [mx-1/2+pxy(:,1), my-1/2+pxy(:,2)]/H;
  end
  
  % determine cell type
  area = sum(int.c{mx,my});
  if area > 0
    cells(mx,my) = fix(area*H^2+1e-4);
  end
end, end

% statistics
if PAR.info
  fprintf('\b\b\b\b     \n inner cells: %4d\n',sum(cells(:)>0)) 
  fprintf(' boundary cells: %4d\n',sum(cells(:)==0)) 
  fprintf(' outer cells: %4d\n',sum(cells(:)<0)) 
end

%%%%% LOCAL FUNCTIONS

function [cxy,pxy] = int_2d(A,H,PAR,GAUSS,POL)
% Gauss coefficients and points in [-1/2,1/2]^2
% subset: sum_i,j A(i,j) x^i y^j > 0

% evaluate at x in {-1/2,1/2} -> 2 edge polynomials in y
Ay = POL.VE*A;

% intersection points on edges in y-direction
y = [-1/2; 1/2];
for k=1:2
   y = [y; real_roots(Ay(k,:))];
end 
y = sort(y); dy = diff(y);
    
% compare error for increasing number of Gauss points
error = inf; 
for step=0:PAR.steps

  % Gauss points for y-intervals
  cy = []; py = [];
  for ky=1:length(dy)
    cy = [cy; dy(ky)*GAUSS{PAR.gauss+step}.C];
    py = [py; y(ky)+dy(ky)*GAUSS{PAR.gauss+step}.P];
  end
  
  % 2d Gauss points
  V = vandermonde(py,PAR.fit); 
  Ay = A*V';
  cxy = []; pxy = [];
  for ky=1:length(py)
     
    % evaluate at y-Gauss-points -> x-polynomials
    [cx,px] = int_1d(Ay(:,ky),PAR,GAUSS);

    % combine cy, py with parameters in x-segment
    cxy = [cxy; cx*cy(ky)];
    pxy = [pxy; [px, ones(size(px,1),1)*py(ky)]]; 
  end

  % compare error for area
  error_new = sum(cxy(:));
  if abs(error-error_new)<PAR.tol_int/(2*H^2)
     break
  end
  error = error_new;
end

%%%%%

function [cx,px] = int_1d(A,PAR,GAUSS)
% Gauss coefficients and points in [-1/2,1/2]^2
% sum_i a_i x^i > 0

% intersections, interval lengths, and signs
x = sort([-1/2; real_roots(A); 1/2]);
dx = diff(x);
s = vandermonde((x(1:end-1)+x(2:end))/2,PAR.fit)*A;

% Gauss points for intervals
cx = []; px = []; 
m_gauss=PAR.gauss+min([1,PAR.steps]);
for kx=1:length(dx)
  if s(kx)>0
    cx = [cx; GAUSS{m_gauss}.C*dx(kx)];
    px = [px; x(kx)+GAUSS{m_gauss}.P*dx(kx)];
  end
end 

% aviod empty vectors for backward compatibility   
if length(cx)==0
  cx = 0; px = 0;
end

%%%%%

function V = vandermonde(t,n)
% (rectangular) Vandermonde matrix v_j,k = t_j^(k-1), k=1:n

V = (t*ones(1,n)).^(ones(length(t),1)*[0:n-1]);

%%%%%

function t = real_roots(p)
% real roots of p_1 + p_2 t + ... in (-1/2,1/2)

t = roots(p(end:-1:1));
t = t(abs(imag(t))<1000*eps & -1/2<t & t<1/2);
t = sort(real(t));

