function [int,cells] = integrate_3d(wD,H,PAR)
% [int,cells] = integrate_3d(wD,H,PAR)
% parameters for numerical integration over grid cells 
%
% function [W,Wx,Wy,Wz] = wD(x,y,z)
%   positive on the domain D, subset of (0,1)^3
% H
%   number of cells per coordinate direction, width 1/H
% PAR
%   algorithmic parameters
% int.c{mx,my,mz}(:)
%   Gauss coefficients for the subset D_m of grid cell m in D
% int.p{mx,my,mz}(:,1:3) 
%   Gauss points for the subset D_m of grid cell m in D
% cells(1:H,1:H,1:H)
%  type of cells; inner=1, boundary=0, outer=-1
% -> see documentation for details 

% authors/copyright 2011: Klaus Hoellig and Joerg Hoerner
% FEMB-Programs (Version: 1.0.1, October 2014)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nintegrate_3d     '); end

% auxiliary variables
load parameters_3d.mat
GAUSS3 = GAUSS3{PAR.gauss}; PTS = PTS{PAR.fit}; POL = POL{PAR.fit};
int.c = cell(H,H,H); int.p = cell(H,H,H); cells = -ones(H,H,H);

% loop over grid cells
work = 0;
for mx=1:H, for my=1:H, for mz=1:H
  work = work+100/H^3;
  if PAR.info, fprintf('\b\b\b\b%3d%%',fix(work)); end
     
  % interpolate values of wD in Q_m on a regular grid X,Y,Z ->
  % Taylor coefficients A with scaled variables in [-1/2,1/2]^3
  A = wD((mx-1+PTS.X)/H,(my-1+PTS.Y)/H,(mz-1+PTS.Z)/H);
  A = reshape(POL.V3*A,PAR.fit,PAR.fit,PAR.fit);

  % dominance of center value: 
  % |2 A(1,1,1)| > Aabs = sum |A(i,j,k)| 2^(-i-j-k+3)
  % -> sufficient criterium to exclude boundary cell 
  Aabs = abs(A(:))'*POL.E; 
  if 2*abs(A(1,1,1)) > Aabs
    if A(1,1,1) > 0 % inner cell 
      int.c{mx,my,mz} = GAUSS3.C/H^3; 
      int.p{mx,my,mz} = [mx-1+GAUSS3.P(:,1), ... 
        my-1+GAUSS3.P(:,2), mz-1+GAUSS3.P(:,3)]/H;
    else % outer cell
      int.c{mx,my,mz} = 0; 
      int.p{mx,my,mz} = [mx-1/2, my-1/2, mz-1/2]/H;
    end
  else

    % order of integration, determined by gradient
    % smallest component last -> outer integration, z-index 
    [g,ind] = sort(-abs([A(2,1,1), A(1,2,1), A(1,1,2)]));
    A = permute(A,ind); 

    % gauss parameters in [-1/2,1/2]^3
    [cxyz,pxyz] = int_3d(A,H,PAR,GAUSS,POL);

    % delete entries with weight 0
    nonzero_ind=find(cxyz>0);
    cxyz=cxyz(nonzero_ind);
    pxyz=pxyz(nonzero_ind,:);
    if isempty(cxyz)
      cxyz=0;pxyz=[0,0,0];
    end        
        
    % reverse permutation due to order of integration
    pxyz(:,ind) = pxyz;
 
    % assign parameters to cell 
    int.c{mx,my,mz} = cxyz/H^3;
    int.p{mx,my,mz} = [mx-1/2+pxyz(:,1), ... 
      my-1/2+pxyz(:,2), mz-1/2+pxyz(:,3)]/H;
  end

  % determine cell type
  area = sum(int.c{mx,my,mz});
  if area > 0 
    cells(mx,my,mz) = fix(area*H^3+PAR.tol_zero);
  end
end, end, end

% statistics
if PAR.info
  fprintf('\b\b\b\b    \n inner cells: %4d\n',sum(cells(:)>0)) 
  fprintf(' boundary cells: %4d\n',sum(cells(:)==0)) 
  fprintf(' outer cells: %4d\n',sum(cells(:)<0)) 
end

%%%%% LOCAL FUNCTIONS

function [cxyz,pxyz] = int_3d(A,H,PAR,GAUSS,POL)
% Gauss weights and points in [-1/2,1/2]^3
% set: sum_ijk a_ijk x^i y^j z^k > 0

% evaluate at x,y in {-1/2,1/2}^2 -> 4 edge polynomials in z
AA = reshape(A,PAR.fit^2,PAR.fit);
Az = POL.VE2*AA;

% intersection points on edges in z-direction
z = [-1/2; 1/2];
for k=1:4;
   z = [z; real_roots(Az(k,:))];
end 
z = sort(z); dz = diff(z);
   
% compare error for increasing number of Gauss points
error = inf; 
for step=0:PAR.steps
  
  % Gauss points for z-intervals
  cz = []; pz = [];
  for kz=1:length(dz)
    cz = [cz; dz(kz)*GAUSS{PAR.gauss+step}.C];
    pz = [pz; z(kz)+dz(kz)*GAUSS{PAR.gauss+step}.P];
  end

  % evaluate at z-Gauss-points -> xy-polynomials 
  V = vandermonde(pz,PAR.fit); 
  Az = AA*V';

  % 3d Gauss points
  cxyz = []; pxyz = [];
  for kz=1:length(pz)
     Azk = reshape(Az(:,kz),PAR.fit,PAR.fit); 
    [cxy,pxy] = int_2d(Azk,H,PAR,GAUSS,POL); 

    % combine cz, pz with parameters in xy-planes 
    cxyz = [cxyz; cxy*cz(kz)];
    pxyz = [pxyz; [pxy, ones(size(pxy,1),1)*pz(kz)]]; 
  end

  % compare error for volume
  error_new = sum(cxyz(:));
  if abs(error-error_new)<PAR.tol_int/(2*H^3)
     break
  end
  error = error_new;
end

%%%%%

function [cxy,pxy] = int_2d(A,H,PAR,GAUSS,POL)
% Gauss weights and points in [-1/2,1/2]^2
% sum_ij a_ij x^i y^j > 0

% evaluate at x in {-1/2,1/2} -> 2 polynomials in y
Ay = POL.VE*A;

% intersection points on edges in y-direction
y = [-1/2; 1/2];
for k=1:2
   y = [y; real_roots(Ay(k,:))];
end 
y = sort(y); dy = diff(y);
   
% compare error for increasing number of Gauss points
error = inf; 
for step=0:PAR.steps
    
  % Gauss points for y-intervals
  cy = []; py = [];
  for ky=1:length(dy)
    cy = [cy; dy(ky)*GAUSS{PAR.gauss+step}.C];
    py = [py; y(ky)+dy(ky)*GAUSS{PAR.gauss+step}.P];
  end
  
  % 2d Gauss points
  V = vandermonde(py,PAR.fit); 
  Ay = A*V';
  cxy = []; pxy = [];
  for ky=1:length(py)
     
    % evaluate at y-Gauss-points -> x-polynomials
    [cx,px] = int_1d(Ay(:,ky),PAR,GAUSS);

    % combine cy, py with parameters in x-segment
    cxy = [cxy; cx*cy(ky)];
    pxy = [pxy; [px, ones(size(px,1),1)*py(ky)]]; 
  end

  % compare error for area
  error_new = sum(cxy(:));
  if abs(error-error_new)<PAR.tol_int/(4*H^2)
     break
  end
  error = error_new;
end

%%%%%

function [cx,px] = int_1d(A,PAR,GAUSS)
% Gauss weights and points in [-1/2,1/2]^2
% sum_i a_i x^i > 0

% intersections, interval lengths, and signs
x = sort([-1/2; real_roots(A); 1/2]);
dx = diff(x);
s = vandermonde((x(1:end-1)+x(2:end))/2,PAR.fit)*A;

% Gauss points for intervals
cx = []; px = []; 
m_gauss = PAR.gauss;+min([1,PAR.steps]);
for k=1:length(dx)
  if s(k)>0
    cx = [cx; GAUSS{m_gauss}.C*dx(k)];
    px = [px; x(k)+GAUSS{m_gauss}.P*dx(k)];
  end
end    

% avoid empty vectors for backward compatibility
if length(cx)==0
  cx = 0; px = 0;
end

%%%%%

function V = vandermonde(t,n)
% Vandermonde matrix v_j,k = t_j^(k-1), k=1:n

V = (t*ones(1,n)).^(ones(length(t),1)*[0:n-1]);

%%%%%

function t = real_roots(p)
% real roots of p_1 + p_2 t + ... in (-1/2,1/2)

t = roots(p(end:-1:1));
t = t(abs(imag(t))<1000*eps & -1/2<t & t<1/2);
t = sort(real(t));

