function U = solve_2d(G,F,PAR)
% U = solve_2d(G,F,PAR)
% conjugate gradient solution of the Ritz-Galerkin system
%
% G(1:H+n,1:H+n,1:2*n+1,1:2*n+1)
%   Ritz-Galerkin matrix in b-spline format
% F(1:H+n,1:H+n)
%   Ritz-Galerkin vector in b-spline format
% PAR
%   algorithmic parameters
% U(1:H+n,1:H+n)
%   scalar coefficients of weighted b-splines
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nsolve_2d      '); end

% initialization
n = (size(G,3)-1)/2;
H = size(G,1)-n; 
norm_F = norm(F(:));

% symmetric diagonal preconditioning
DG = ones(H+3*n,H+3*n); 
ind = [1:H+n];
DG(n+ind,n+ind) = 1./sqrt(squeeze(G(:,:,n+1,n+1)));
for sx=1:2*n+1, for sy=1:2*n+1
  G(:,:,sx,sy) = G(:,:,sx,sy).*DG(n+ind,n+ind).* ... 
    DG(sx-1+ind,sy-1+ind);
end, end
F = F.*DG(n+ind,n+ind);

% start values
U = zeros(H+n,H+n);
D = -F; V = -D; W = 0*D;
gamma = D(:)'*D(:); 
tol = gamma*PAR.tol_solve^2;

% iteration 
for steps = 1:PAR.iterations
  if PAR.info & mod(steps,10)==0, fprintf('\b\b\b\b\b\b%6d',steps), end      
   W = mult(G,V,H,n);
   alpha = gamma/(V(:)'*W(:)); beta = 1/gamma;
   U = U + alpha*V;
   D = D + alpha*W; 
   gamma = D(:)'*D(:);
   if gamma < tol
     break
   end
   beta = beta*gamma; 
   V = -D + beta*V;
end 

% error 
error = (mult(G,U,H,n)-F)./DG(n+ind,n+ind);
if PAR.info
  fprintf('\b\b\b\b\b\b      \n unknowns: %4d\n',(H+n)^2) 
  fprintf(' iterations: %4d\n',steps) 
  fprintf(' error of ls: %10.3e\n',norm(error(:))/norm_F);
end

% reverse preconditioning 
U = U.*DG(n+ind,n+ind);

%%%%% LOCAL FUNCTION 

function W = mult(G,V,H,n)
% sparse matrix/vector multiplication

V0 = zeros(H+3*n,H+3*n); 
V0(n+1:H+2*n,n+1:H+2*n) = V;
W = zeros(H+n,H+n);
for sx=1:2*n+1, for sy=1:2*n+1
  W = W + G(:,:,sx,sy).*V0(sx:end-2*n-1+sx,sy:end-2*n-1+sy); 
end, end

