function U = solve_3de(G,F,PAR)
% U = solve_3de(G,F,PAR)
% conjugate gradient solution of Ritz-Galerkin block system
%
% G(1:H+n,1:H+n,1:H+n,1:2*n+1,1:2*n+1,1:2*n+1,1:3,1:3)
%   Ritz-Galerkin matrix in b-spline format
% F(1:H+n,1:H+n,1:H+n,1:3)
%   Ritz-Galerkin vector in b-spline format
% PAR
%   algorithmic parameters
% U(1:H+n,1:H+n,1:H+n,1:3)
%   vector coefficients of weighted b-splines
% -> see documentation for details 

% FEMB-Programs (Version: 1.0)
% test version: for educational use and illustration purposes only
% -> see documentation for details

if PAR.info, fprintf('\nsolve_3de      '); end

% initialization
n = (size(G,4)-1)/2;
H = size(G,1)-n; 
Hn = H+n; nn = 2*n+1;

% symmetric diagonal preconditioning
DG = ones(H+3*n,H+3*n,H+3*n,3);
ind = [1:Hn];
for m=1:3 
  DG(n+ind,n+ind,n+ind,m) = ... 
    1./sqrt(squeeze(G(:,:,:,n+1,n+1,n+1,m,m)));
end
for sx=1:nn, for sy=1:nn, for sz=1:nn, for m1=1:3, for m2=1:3
  G(:,:,:,sx,sy,sz,m1,m2) = G(:,:,:,sx,sy,sz,m1,m2) .* ... 
    DG(n+ind,n+ind,n+ind,m1).*DG(sx-1+ind,sy-1+ind,sz-1+ind,m2);
end, end, end, end, end
F = F.*DG(n+ind,n+ind,n+ind,:);

% start values
U = zeros(Hn,Hn,Hn,3);
D = -F; V = -D; W = 0*D;
gamma = D(:)'*D(:); 
tol = gamma*PAR.tol_solve^2;

% iteration 
for steps = 1:PAR.iterations
  if PAR.info & mod(steps,10)==0, fprintf('\b\b\b\b\b\b%6d',steps), end   
  W = mult(G,V,H,n);
  alpha = gamma/(V(:)'*W(:)); beta = 1/gamma;
  U = U + alpha*V;
  D = D + alpha*W; 
  gamma = D(:)'*D(:);
  if gamma < tol
    break
  end
  beta = beta*gamma; 
  V = -D + beta*V;
end 

% error 
fprintf('\b\b\b\b    \n unknowns: %4d\n',Hn^3) 
fprintf(' iterations: %4d\n',steps) 
error = (mult(G,U,H,n)-F)./DG(n+ind,n+ind,n+ind,:);
fprintf(' error of ls: %10.3e\n',norm(error(:),inf)) 

% reverse preconditioning
U = U.*DG(n+ind,n+ind,n+ind,:);

%%%%% LOCAL FUNCTION 

function W = mult(G,V,H,n)
% sparse matrix/vector multiplication

Hn = H+n; nn = 2*n+1;
ind = [1:Hn];
V0 = zeros(H+3*n,H+3*n,H+3*n,3); 
V0(n+ind,n+ind,n+ind,:) = V;
W = zeros(Hn,Hn,Hn,3);

for sx=1:nn, for sy=1:nn, for sz=1:nn, for m=1:3
  W(:,:,:,m) = W(:,:,:,m) + ... 
    G(:,:,:,sx,sy,sz,m,1) .* ... 
    V0(sx:end-nn+sx,sy:end-nn+sy,sz:end-nn+sz,1) + ...
    G(:,:,:,sx,sy,sz,m,2) .* ... 
    V0(sx:end-nn+sx,sy:end-nn+sy,sz:end-nn+sz,2) + ...
    G(:,:,:,sx,sy,sz,m,3) .* ... 
    V0(sx:end-nn+sx,sy:end-nn+sy,sz:end-nn+sz,3);
end, end, end, end

